'''Remove files.
'''

import os, stat, shutil

def remove_dir_with_permissions(dir_path):

    """Remove temporary files.

    Args:
        dir_path (str): path to your file.

    """
    def handle_permission_error(func, path, exc_info):
        os.chmod(path, stat.S_IWUSR)
        func(path)

    if os.path.exists(dir_path):
        shutil.rmtree(dir_path, onerror=handle_permission_error)


"""Analysis topology, open metal sites, revised autocorrelation and so on.
"""

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.ase import AseAtomsAdaptor
from ase.io import read
from pymatgen.core.structure import Structure

import os,juliacall
from CoREMOF.calculation.mof_collection import MofCollection
from CoREMOF.utils import remove

from molSimplify.Informatics.MOF.MOF_descriptors import get_MOF_descriptors


def SpaceGroup(structure):

    """Analysis space group of structure.

    Args:
        structure (str): path to your CIF.
       
    Returns:
        Dictionary:
            -   unit by ["unit"], always nan
            -   hall symbol by ["hall_symbol"]
            -   space group number by ["space_group_number"]
            -   crystal system by ["crystal_system"]
    """
           
    result_sg = {}
    result_sg["unit"]="nan"
    atoms = read(structure)
    structure_ = AseAtomsAdaptor.get_structure(atoms)
    result_ = SpacegroupAnalyzer(structure_, symprec=0.01, angle_tolerance=5)
    hall_symbol = result_.get_hall()
    space_group_number = result_.get_space_group_number()
    crystal_system = result_.get_crystal_system()
    result_sg["hall_symbol"]=hall_symbol
    result_sg["space_group_number"]=space_group_number
    result_sg["crystal_system"]=crystal_system

    return result_sg

def Mass(structure):

    """Analysis total mass of structure.

    Args:
        structure (str): path to your CIF.
       
    Returns:
        Dictionary:
            -   unit by ["unit"], always amu
            -   total mass by ["total_mass"]
    """

    result_m = {}
    result_m["unit"]="amu"
    atoms = read(structure)
    total_mass = atoms.get_masses().sum()
    result_m["total_mass"]=float(total_mass)

    return result_m

def Volume(structure):

    """Analysis total volume of structure.

    Args:
        structure (str): path to your CIF.
       
    Returns:
        Dictionary:
            -   unit by ["unit"], always Å^3
            -   total volume by ["total_volume"]
    """

    result_v = {}
    result_v["unit"]="Å^3"
    with open(structure, "r", encoding="utf-8") as f:
        cif_file = f.read()
    atoms = Structure.from_str(cif_file, fmt="cif")
    total_volume = atoms.volume
    result_v["total_volume"]=total_volume

    return result_v

def n_atom(structure):

    """Analysis number of atoms of structure.

    Args:
        structure (str): path to your CIF.
       
    Returns:
        Dictionary:
            -   unit by ["unit"], always nan
            -   number of atoms by ["number_atoms"]
    """

    result_na = {}
    result_na["unit"]="nan"
    atoms = read(structure)
    number_atoms = len(atoms)
    result_na["number_atoms"]=number_atoms

    return result_na

def topology(structure, node_type="single"):

    """Analysis topology of structure by CrystalNets.jl (https://github.com/coudertlab/CrystalNets.jl?tab=readme-ov-file).

    Args:
        structure (str): path to your CIF.
        node_type (str): the clustering algorithm used to group atoms into vertices. single: each already-defined cluster is mapped to a vertex; all: keep points of extension for organic clusters.
       
    Returns:
        Dictionary:
            -   dimension by ["dimension"]
            -   topology by ["topology"]
            -   catenation by ["catenation"]
    """

    package_directory = os.path.abspath(__file__).replace("mof_features.py","")
    os.environ["JULIA_DEPOT_PATH"] = package_directory

    juliacall.Main.seval('import Pkg; Pkg.add("CrystalNets")')

    jl = juliacall.newmodule("topo")
    jl.seval("using CrystalNets")

    if node_type == "single":
        clustering = jl.Clustering.SingleNodes
    elif node_type == "all":
        clustering = jl.Clustering.AllNodes
    else:
        raise ValueError("node_type should be single or all")
    options = jl.CrystalNets.Options(structure=jl.StructureType.MOF, clusterings=[clustering])
    result = jl.determine_topology(structure, options)
    
    result_tp = {}
    result_tp["dimension"] = []
    result_tp["topology"] = []
    result_tp["catenation"] = []
    interpenetration = jl.CrystalNets.total_interpenetration(result, clustering)
    for x in result:
        info = x[0][clustering]
        result_tp["dimension"].append(jl.ndims(info.genome))
        result_tp["topology"].append(str(info))
        result_tp["catenation"].append(interpenetration[info])

    return result_tp


def get_oms_file(structure):

    """Analysis open metal site of structure from CoRE MOF 2019 (https://github.com/emmhald/open_metal_detector).

    Args:
        structure (str): path to your CIF.
       
    Returns:
        Dictionary:
            -   all types of metal by ["Metal Types"]
            -    has OMS or not by ["Has OMS"], -> True or False
            -    of type of OMS if has by ["OMS Types"]
    """
        
    a_mof_collection = MofCollection(path_list = [structure], 
                                 analysis_folder="tmp_oms")
    a_mof_collection.analyse_mofs(num_batches=1,overwrite=False)
    oms_result = {
                    "Metal Types": a_mof_collection.mof_oms_df["Metal Types"][structure.replace(".cif","").split("/")[-1]],
                    "Has OMS": a_mof_collection.mof_oms_df["Has OMS"][structure.replace(".cif","").split("/")[-1]],
                    "OMS Types": a_mof_collection.mof_oms_df["OMS Types"][structure.replace(".cif","").split("/")[-1]],
                }

    remove.remove_dir_with_permissions("tmp_oms")

    return oms_result


def get_oms_folder(input_folder, n_batch = 1):

    """Analysis open metal site of folder with structures from CoRE MOF 2019 (https://github.com/emmhald/open_metal_detector).

    Args:
        input_folder (str): path to your folder.
        n_batch (int): number of batches.
       
    Returns:
        Dictionary:
            -   all types of metal of each structure by [structure]["Metal Types"]
            -   has OMS or not of each structure by [structure]["Has OMS"], -> True or False
            -   type of OMS if has of each structure by [structure]["OMS Types"]
    """

    mof_collection = MofCollection.from_folder(collection_folder = input_folder, 
                                                analysis_folder="tmp_oms")
    mof_collection.analyse_mofs(num_batches=n_batch,overwrite=False)
    oms_result = {}
    for name, row in mof_collection.mof_oms_df.iterrows():
        
        oms_result[name] = {
                            "Metal Types": row[0],
                            "Has OMS": row[1],
                            "OMS Types": row[2]
                        }

    remove.remove_dir_with_permissions("tmp_oms")

    return oms_result


def RACs(structure):

    """Revised Autocorrelation features (https://github.com/hjkgrp/molSimplify).

    Args:
        input_folder (str): path to your folder.
        n_batch (int): number of batches.
       
    Returns:
        Dictionary:
            -   metal by ["Metal"]
            -   linker by ["Linker"]
            -   function group by ["Function-group"]
    """

    metal_fnames = ['D_mc-I-0-all', 'D_mc-I-1-all', 'D_mc-I-2-all', 'D_mc-I-3-all', 'D_mc-S-0-all', 'D_mc-S-1-all', 'D_mc-S-2-all',
    'D_mc-S-3-all', 'D_mc-T-0-all', 'D_mc-T-1-all', 'D_mc-T-2-all', 'D_mc-T-3-all', 'D_mc-Z-0-all', 'D_mc-Z-1-all', 'D_mc-Z-2-all',
    'D_mc-Z-3-all', 'D_mc-chi-0-all', 'D_mc-chi-1-all', 'D_mc-chi-2-all', 'D_mc-chi-3-all', 'f-I-0-all', 'f-I-1-all', 'f-I-2-all',
    'f-I-3-all', 'f-S-0-all', 'f-S-1-all', 'f-S-2-all', 'f-S-3-all', 'f-T-0-all', 'f-T-1-all', 'f-T-2-all', 'f-T-3-all', 'f-Z-0-all',
    'f-Z-1-all', 'f-Z-2-all', 'f-Z-3-all', 'f-chi-0-all', 'f-chi-1-all', 'f-chi-2-all', 'f-chi-3-all', 'mc-I-0-all', 'mc-I-1-all',
    'mc-I-2-all', 'mc-I-3-all', 'mc-S-0-all', 'mc-S-1-all', 'mc-S-2-all', 'mc-S-3-all', 'mc-T-0-all', 'mc-T-1-all', 'mc-T-2-all',
    'mc-T-3-all', 'mc-Z-0-all', 'mc-Z-1-all', 'mc-Z-2-all', 'mc-Z-3-all', 'mc-chi-0-all', 'mc-chi-1-all', 'mc-chi-2-all', 'mc-chi-3-all']

    linker_fnames = ['D_lc-I-0-all', 'D_lc-I-1-all', 'D_lc-I-2-all', 'D_lc-I-3-all', 'D_lc-S-0-all', 'D_lc-S-1-all', 'D_lc-S-2-all',
    'D_lc-S-3-all', 'D_lc-T-0-all', 'D_lc-T-1-all', 'D_lc-T-2-all', 'D_lc-T-3-all', 'D_lc-Z-0-all', 'D_lc-Z-1-all', 'D_lc-Z-2-all',
    'D_lc-Z-3-all', 'D_lc-alpha-0-all', 'D_lc-alpha-1-all', 'D_lc-alpha-2-all', 'D_lc-alpha-3-all', 'D_lc-chi-0-all', 'D_lc-chi-1-all',
    'D_lc-chi-2-all', 'D_lc-chi-3-all', 'lc-I-0-all', 'lc-I-1-all', 'lc-I-2-all', 'lc-I-3-all', 'lc-S-0-all', 'lc-S-1-all', 'lc-S-2-all',
    'lc-S-3-all', 'lc-T-0-all', 'lc-T-1-all', 'lc-T-2-all', 'lc-T-3-all', 'lc-Z-0-all', 'lc-Z-1-all', 'lc-Z-2-all', 'lc-Z-3-all',
    'lc-alpha-0-all', 'lc-alpha-1-all', 'lc-alpha-2-all', 'lc-alpha-3-all', 'lc-chi-0-all', 'lc-chi-1-all', 'lc-chi-2-all', 'lc-chi-3-all',
    'f-lig-I-0', 'f-lig-I-1', 'f-lig-I-2', 'f-lig-I-3', 'f-lig-S-0', 'f-lig-S-1', 'f-lig-S-2', 'f-lig-S-3', 'f-lig-T-0', 'f-lig-T-1',
    'f-lig-T-2', 'f-lig-T-3', 'f-lig-Z-0', 'f-lig-Z-1', 'f-lig-Z-2', 'f-lig-Z-3', 'f-lig-chi-0', 'f-lig-chi-1', 'f-lig-chi-2', 'f-lig-chi-3']

    fg_fnames = ['D_func-I-0-all', 'D_func-I-1-all', 'D_func-I-2-all', 'D_func-I-3-all', 'D_func-S-0-all', 'D_func-S-1-all',
    'D_func-S-2-all', 'D_func-S-3-all', 'D_func-T-0-all', 'D_func-T-1-all', 'D_func-T-2-all', 'D_func-T-3-all', 'D_func-Z-0-all',
    'D_func-Z-1-all', 'D_func-Z-2-all', 'D_func-Z-3-all', 'D_func-alpha-0-all', 'D_func-alpha-1-all', 'D_func-alpha-2-all',
    'D_func-alpha-3-all', 'D_func-chi-0-all', 'D_func-chi-1-all', 'D_func-chi-2-all', 'D_func-chi-3-all', 'func-I-0-all',
    'func-I-1-all', 'func-I-2-all', 'func-I-3-all', 'func-S-0-all', 'func-S-1-all', 'func-S-2-all', 'func-S-3-all', 'func-T-0-all',
    'func-T-1-all', 'func-T-2-all', 'func-T-3-all', 'func-Z-0-all', 'func-Z-1-all', 'func-Z-2-all', 'func-Z-3-all', 'func-alpha-0-all',
    'func-alpha-1-all', 'func-alpha-2-all', 'func-alpha-3-all', 'func-chi-0-all', 'func-chi-1-all', 'func-chi-2-all', 'func-chi-3-all']

    result_rac = {}
    result_rac["Metal"] = {}
    result_rac["Linker"] = {}
    result_rac["Function-group"] = {}

    os.makedirs("tmp_rac", exist_ok=True)
    
    name = os.path.basename(structure).replace(".cif", "")
    
    full_names, full_descriptors = get_MOF_descriptors(
                                                        structure,
                                                        3,
                                                        path='tmp_rac',
                                                        xyzpath=f'tmp_rac/{name}.xyz',
                                                        max_num_atoms=6000
                                                        )
                                                        
    descriptor_data = dict(zip(full_names, full_descriptors))

    for metal in metal_fnames:
        result_rac["Metal"][metal] =  round(float(descriptor_data[metal]), 4)
    for linker in linker_fnames:
        result_rac["Linker"][linker] =  round(float(descriptor_data[linker]), 4)
    for fg in fg_fnames:
        result_rac["Function-group"][fg] =  round(float(descriptor_data[fg]), 4)

    remove.remove_dir_with_permissions("tmp_rac")

    return result_rac



"""Process your CIF to get mofid v1 and v2.
"""

from ase.io import read as ase_read
from ase.io import write as ase_write
from ase import neighborlist
import networkx as nx
from ase.build import sort
import ase
import os, glob, shutil
import numpy as np
import collections

from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice

from mofid.id_constructor import extract_fragments
from collections import Counter
from mofid.run_mofid import cif2mofid

from pymatgen.analysis.structure_matcher import ElementComparator, StructureMatcher
import selfies as sf

from openbabel import openbabel as ob


metals  = ['Li', 'Na', 'K', 'Rb', 'Cs', 'Fr', 'Be', 'Mg', 'Ca', 'Sr', 'Ba', 'Ra',
          'Al', 'Ga', 'Ge', 'In', 'Sn', 'Sb', 'Tl', 'Pb', 'Bi', 'Po',
          'Sc', 'Ti', 'V' , 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
          'Y' , 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
          'Hf', 'Ta', 'W' , 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',
          'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'U', 'Tm', 'Yb', 'Lu',
          'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr'
         ]

def run_v1(structure):

    """Converting CIF to mofid-v1, see https://snurr-group.github.io/mofid/ for additional installation. Tip: Please check CMAKE, JAVA, etc. before installing.

    Args:
        structure (str): path to your CIF.

    Returns:
        String:
            -   mofid-v1.
    """

    mofid_v1 = cif2mofid(structure)
    return mofid_v1 # mofid_v1['mofid'], mofid_v1['smiles_nodes'], mofid_v1['smiles_linkers'], mofid_v1['topology'], mofid_v1['cat']

def dict2str(dct):
    """Convert symbol-to-number dict to str.
    """
    return ''.join(symb + (str(n)) for symb, n in dct.items())

def split_nodes_from_cif(structure, prefix='Output'):
    """Split nodes (CIF) to single XYZ.

    Args:
        structure (str): path to your CIF.
        prefix (str): the path to save processed XYZ.

    Returns:
        int:
            -   0: sucess; 1: fail.
    """
    try:
        atoms = ase_read(structure)
    except:
        print('Error with reading CIF in {}'.format(structure))
        return 1
    cutOff = neighborlist.natural_cutoffs(atoms)
    neighborList = neighborlist.NeighborList(cutOff, self_interaction=False, bothways=True, skin=0.3)
    neighborList.update(atoms)
    G = nx.Graph()
    for k in range(len(atoms)):
        tup = (k, {"element":"{}".format(atoms.get_chemical_symbols()[k]), "pos": atoms.get_positions()[k]})
        G.add_nodes_from([tup])
    for k in range(len(atoms)):
        for i in neighborList.get_neighbors(k)[0]:
            G.add_edge(k, i)
    Gcc = sorted(nx.connected_components(G), key=len, reverse=True)
    form_dicts = []
    for index, g in enumerate(Gcc):
        g = list(g)
        fragment = atoms[g]
        fragment = sort(fragment)
        form_dict = fragment.symbols.formula.count()
        form_dicts.append(dict2str(form_dict))
    #nodes = [atoms[list(Gcc[0])]]
    #unique_formdicts = [form_dicts[0]]
    nodes = []
    unique_formdicts = []
    if len(form_dicts) > 1:
        for index, form_dict in enumerate(form_dicts):
            if form_dict not in unique_formdicts:
                nodes.append(atoms[list(Gcc[index])])
                unique_formdicts.append(form_dict)
    elif len(form_dicts) == 1:
        nodes.append(atoms[list(Gcc[0])])
        unique_formdicts.append(form_dicts[0])
    for index, _ in enumerate(nodes):
        ase_write('{}/node{}.xyz'.format(prefix, index), nodes[index]) # /AllNode/nodes.cif
    return 0

def split_linkers_from_cif(structure, prefix='Output'):
    """Split linkers (CIF) to single XYZ.

    Args:
        structure (str): path to your CIF.
        prefix (str): the path to save processed XYZ.

    Returns:
        int:
            -   0: sucess; 1: fail.
    """
    try:
        atoms = ase_read(structure)
    except:
        print('Error with reading CIF in {}'.format(structure))
        return 1
    cutOff = neighborlist.natural_cutoffs(atoms)
    neighborList = neighborlist.NeighborList(cutOff, self_interaction=False, bothways=True, skin=0.3)
    neighborList.update(atoms)
    G = nx.Graph()
    for k in range(len(atoms)):
        tup = (k, {"element":"{}".format(atoms.get_chemical_symbols()[k]), "pos": atoms.get_positions()[k]})
        G.add_nodes_from([tup])
    for k in range(len(atoms)):
        for i in neighborList.get_neighbors(k)[0]:
            G.add_edge(k, i)
    Gcc = sorted(nx.connected_components(G), key=len, reverse=True)
    form_dicts = []
    for index, g in enumerate(Gcc):
        g = list(g)
        fragment = atoms[g]
        fragment = sort(fragment)
        form_dict = fragment.symbols.formula.count()
        form_dicts.append(dict2str(form_dict))
    nodes = []
    unique_formdicts = []
    if len(form_dicts) > 1:
        for index, form_dict in enumerate(form_dicts):
            if form_dict not in unique_formdicts:
                nodes.append(atoms[list(Gcc[index])])
                unique_formdicts.append(form_dict)
    elif len(form_dicts) == 1:
        nodes.append(atoms[list(Gcc[0])])
        unique_formdicts.append(form_dicts[0])
    for index, _ in enumerate(nodes):
        ase_write('{}_MetalOxolinker{}.xyz'.format(prefix, index), nodes[index]) # /MetalOxo/linkers.cif
    return 0

def get_node_linker_files(structure, prefix='Output'):
    """Split MOF to sbu + linker.

    Args:
        structure (str): path to your CIF.
        prefix (str): the path to save processed XYZ.

    Returns:
        files:
            -   structure, node, linker, ...
    """
    os.makedirs(prefix, exist_ok=True)
    extract_fragments(structure, prefix)

def xyz2fomula(xyzpath):
    """from XYZ to chemical fomula based on A-Z.

    Args:
        xyzpath (str): path to your XYZ.

    Returns:
        str:
            -   fomula.
    """
    atoms = ase_read(xyzpath)
    symbols = atoms.get_chemical_symbols()
    counts = Counter(symbols)
    sorted_items = sorted(counts.items(), key=lambda x: x[0])
    # formula = ''.join(f"{el}{n if n > 1 else ''}" for el, n in sorted_items)
    formula = ''.join(f"{el}{n}" for el, n in sorted_items)
    return formula

def convert_ase_pymat(ase_objects):
    """convert to ase to pymatgen atoms.

    Args:
        ase_objects (ase.Atoms): ase-type atoms.

    Returns:
        ase.Atoms:
            -   pymatgen-type atoms.
    """
    structure_lattice = Lattice(ase_objects.cell)
    structure_species = ase_objects.get_chemical_symbols()
    structure_positions = ase_objects.get_positions()
    return Structure(structure_lattice,structure_species,structure_positions)

def remove_pbc_cuts(atoms):
    """Remove building block cuts due to periodic boundary conditions. After the
    removal, the atoms object is centered at the center of the unit cell.

    Args:
        atoms (ase.Atoms): aase.Atoms.

    Returns:
        ase.Atoms:
            -   The processed atoms object.
    """

    try:
        # setting cuttoff parameter
        scale = 1.4
        cutoffs = ase.neighborlist.natural_cutoffs(atoms)
        cutoffs = [scale * c for c in cutoffs]
        #making neighbor_list
        #graphs of single metal is not constructed because there is no neighbor.
        I, J, D = ase.neighborlist.neighbor_list("ijD",atoms,cutoff=cutoffs)
        nl = [[] for _ in atoms]
        for i, j, d in zip(I, J, D):
            nl[i].append((j, d))
        visited = [False for _ in atoms]
        q = collections.deque()
        # Center of the unit cell.
        abc_half = np.sum(atoms.get_cell(), axis=0) * 0.5
        positions = {}
        q.append((0, np.array([0.0, 0.0, 0.0])))
        while q:
            i, pos = q.pop()
            visited[i] = True
            positions[i] = pos
            for j, d in nl[i]:
                if not visited[j]:
                    q.append((j, pos + d))
                    visited[j] = True
        centroid = np.array([0.0, 0.0, 0.0])
        for v in positions.values():
            centroid += v
        centroid /= len(positions)
        syms = [None for _ in atoms]
        poss = [None for _ in atoms]
        for i in range(len(atoms)):
            syms[i] = atoms.symbols[i]
            poss[i] = positions[i] - centroid + abc_half
        atoms = ase.Atoms(
            symbols=syms, positions=poss, pbc=True, cell=atoms.get_cell()
        )
        # resize of cell
        cell_x = np.max(atoms.positions[:,0]) - np.min(atoms.positions[:,0])
        cell_y = np.max(atoms.positions[:,1]) - np.min(atoms.positions[:,1])
        cell_z = np.max(atoms.positions[:,2]) - np.min(atoms.positions[:,2])
        cell = max([cell_x,cell_y,cell_z])
        atoms.set_cell([cell+2,cell+2,cell+2, 90,90,90])
        center_mass = atoms.get_center_of_mass()
        cell_half  = atoms.cell.cellpar()[0:3]/2
        atoms.positions = atoms.positions - center_mass + cell_half    
        return atoms
    except:
        return atoms

def run_v2(structure, nodes_dataset, refname):
    """run mofidv2 from CIF.

    Args:
        structure (str): path to your MOF.
        nodes_dataset (str): path to your node dataset (download from https://github.com/sxm13/CoREMOF_tools/tree/main/CoREMOF/data/mofid/nodes.zip).
        refname (str): file name of your structure or define by yourself.

    Returns:
        str:
            -   mofid-v2.
    """
    # get list of nodes
    try:
        shutil.rmtree("Output")
    except:
        pass
    # get list of nodes
    nodes_type = glob.glob(nodes_dataset+"/*xyz")
    nodes = []
    for node_file in nodes_type:
        nodes.append(node_file.split("/")[-1].split("_")[0])
    nodes = list(set(nodes))
    # get information of mofid-v1
    mofidv1 = run_v1(structure)
    linkers = mofidv1["smiles_linkers"]
    all_linkers = []
    for linker in linkers:
        all_linkers.append(sf.encoder(linker))
    topology = mofidv1["topology"]
    cat = mofidv1["cat"]
    # get_node_linker_files(cifpath)
    check = split_nodes_from_cif("Output/AllNode/nodes.cif", "Output")
    if check == 1:
        print("nan")
        return "nan"
    else:
        all_nodes_xyz = glob.glob("./Output/node*xyz")
        all_nodes_part = []
        for node_xyz in all_nodes_xyz:
            node_formula = xyz2fomula(node_xyz)
            if node_formula in nodes:
                matcher = StructureMatcher(ltol = 0.3,
                                           stol = 2,
                                           angle_tol = 5,
                                           primitive_cell = False,
                                           scale = False,
                                           comparator=ElementComparator()) 
                matched = False
                known_nodes = glob.glob(nodes_dataset+ "/" + node_formula + "*xyz")
                
                mof = remove_pbc_cuts(ase_read(node_xyz))
                a = convert_ase_pymat(mof)
                for i in range(len(known_nodes)):
                    b = convert_ase_pymat(remove_pbc_cuts(ase_read(known_nodes[i])))
                    if matcher.fit(a, b):
                        node_part = os.path.basename(known_nodes[i].replace(".xyz", ""))
                        all_nodes_part.append(node_part)
                        matched = True
                        print(node_formula, "the node can be found in nodes dataset")
                        break
                if not matched:
                    # raise RuntimeError(f"fail matched from nodes dataset, stop")
                    node_part = node_formula + "_Type-" + str(len(known_nodes) + 1)
                    all_nodes_part.append(node_part)
                    shutil.move(node_xyz, nodes_dataset + "/" + node_part + ".xyz")
                    print("new node found, has moved the nodes dataset")
            else:
                node_part = node_formula + "_Type-1"
                all_nodes_part.append(node_part)
                shutil.move(node_xyz, nodes_dataset + "/" + node_part + ".xyz")
                print("new node found, has moved the nodes dataset")

    linkers_part = ".".join(all_linkers)
    nodes_part = ".".join(f"[{node}]" for node in all_nodes_part)
    mofidv2 = nodes_part + "." + linkers_part + " " + "MOFid-v2." + topology + ".cat" + cat + ";" + refname
    try:
        shutil.rmtree("Output")
    except:
        pass
    return mofidv2

def are_identical_smiles(smiles1, smiles2):
    obConversion = ob.OBConversion()
    obConversion.SetInFormat("smi")
    obConversion.SetOutFormat("can")  # Set output to canonical SMILES
    mol1 = ob.OBMol()
    mol2 = ob.OBMol()
    # Read the SMILES strings into molecules
    obConversion.ReadString(mol1, smiles1)
    obConversion.ReadString(mol2, smiles2)
    # Convert molecules to canonical SMILES
    can_smiles1 = obConversion.WriteString(mol1).strip()
    can_smiles2 = obConversion.WriteString(mol2).strip()
    # Compare canonical SMILES
    return can_smiles1 == can_smiles2



"""Download structures and query information of CoRE MOF Database.
"""

try:
    from ccdc import io
    csd_reader = io.EntryReader('CSD')
except:
    print("You need to install CSD software with the license if you want to download all structures")

import os, json, requests, zipfile

from gemmi import cif

package_directory = os.path.abspath(__file__).replace("structure.py","")

files_to_download = {
                    'data/CR.json': 'https://raw.githubusercontent.com/sxm13/CoREMOF_tools/main/CoREMOF/data/info/CR.json',
                    'data/NCR.json': 'https://raw.githubusercontent.com/sxm13/CoREMOF_tools/main/CoREMOF/data/info/NCR.json',
                    'data/SI/CR.zip': 'https://raw.githubusercontent.com/sxm13/CoREMOF_tools/main/CoREMOF/data/SI/CR.zip',
                    'data/SI/NCR.zip': 'https://raw.githubusercontent.com/sxm13/CoREMOF_tools/main/CoREMOF/data/SI/NCR.zip'
                    }

for file_name, url in files_to_download.items():
    
    file_path = os.path.join(package_directory, file_name)
    directory = os.path.dirname(file_path) 

    if not os.path.exists(directory):
        os.makedirs(directory)
        
    if not os.path.exists(file_path):
        response = requests.get(url)
        if response.status_code == 200:
            with open(file_path, 'wb') as f:
                f.write(response.content)
            print(f"Downloaded {file_name} to {file_path}")
        else:
            print(f"Failed to download {file_name} from {url}")
    else:
        pass

class download_from_SI():

    """download structures that we got from supporting information.

    Args:
        output_folder (str): path to save structures.

    Returns:
        cif:
            CoRE MOF SI dataset.   
    """

    def __init__(self, output_folder="./CoREMOF2024DB"):
        
        self.SI_path = package_directory+'/data/SI/'
        self.output = output_folder
        self.run()

    def run(self):

        """start to run. 
        """
            
        CR_files = self.list_zip(self.SI_path+"CR.zip")
        NCR_files = self.list_zip(self.SI_path+"NCR.zip")
     
        os.makedirs(self.output+"/CR/", exist_ok=True)
        os.makedirs(self.output+"/NCR/", exist_ok=True)

        for file in CR_files[:]:
            self.get_from_SI(self.SI_path+"CR.zip", file, self.output)
        for file in NCR_files[:]:
            self.get_from_SI(self.SI_path+"NCR.zip", file, self.output)

    def list_zip(self, zip_path):

        """list of files from a ZIP.

        Args:
            zip_path (str): path to ZIP.

        Returns:
            List:
                name list from a ZIP.  
        """
            
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            file_list = zip_ref.namelist()
            return file_list
    
    def get_from_SI(self, zip_path, entry, output_folder):

        """unzip files from a ZIP.

        Args:
            zip_path (str): path to ZIP.
            entry (str): name of structure.
            output_folder (str): path to save structures. 
        """
                
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            file_list = zip_ref.namelist()
            if entry in file_list:
                zip_ref.extract(entry, output_folder)
            
def download_from_CSD(refcode, output_folder="./CoREMOF2024DB"):

    """download structures from CSD, you need to install [CSD python API](https://downloads.ccdc.cam.ac.uk/documentation/API/installation_notes.html) with licence.

    Args:
        refcode (str): CSD refcode.
        output_folder (str): path to save structures.

    Returns:
        cif:
            downloading CIF.  
    """

    cryst = csd_reader.crystal(refcode)
    data = cryst.to_string('cif')
    f = open(os.path.join(output_folder, refcode+'.cif'),'w')
    f.write(data)
    f.close()


def information(dataset, entry):

    """get information of CoRE MOF database.

    Args:
        dataset (str): name of subset.
        entry (str): name of structure

    Returns:
        Dictionary:
            properties, DOI, issues and so on. 
    """     

    CR_data_path = package_directory+'/data/CR.json'
    NCR_data_path = package_directory+'/data/NCR.json'

    with open (CR_data_path, "r") as CR_f:
        CR_data = json.load(CR_f)
        
    with open (NCR_data_path, "r") as NCR_f:
        NR_data = json.load(NCR_f)
    
    if dataset == "CR-ASR":
        print("unit:\n", CR_data["unit"])
        query_data = CR_data["ASR"]
    elif dataset == "CR-FSR":
        print("unit:\n", CR_data["unit"])
        query_data = CR_data["FSR"]
    elif dataset == "CR-Ion":
        print("unit:\n", CR_data["unit"])
        query_data = CR_data["Ion"]
    elif dataset == "NCR":
        query_data = NR_data

    return query_data[entry]

def read_aif(GEMC_data):

    """get adsorption amount of water from GEMC.

    Args:
        GEMC_data (list): from detail_of_CR.json, for example, information("CR-ASR", "2020[Cu][sql]2[ASR]1")["GEMC"].

    Returns:
        Dictionary:
            -   information,by ["info"] always "('_units_loading', 'Molecules/Supercell')".
            -   pressure by ["pressure"].
            -   uptake by ["uptake"].
    """    
        
    with open("temp_gemc.aif", "w") as f:
        f.write("".join(GEMC_data) )
    data = cif.read_file("temp_gemc.aif")
    os.remove("temp_gemc.aif")
    block = data.sole_block()

    item = block.find_pair_item('_units_loading')

    adsorption_data = {}
    adsorption_data["info"]=item.pair
    adsorption_data["pressure"]=list(block.find_loop('_adsorp_pressure'))
    adsorption_data["uptake"]=list(block.find_loop('_adsorp_amount'))

    return adsorption_data



"""Geometery properties calculation based on Zeo++ software
You need install Zeo++ package independently from source (https://www.zeoplusplus.org/download.html) or conda (https://anaconda.org/conda-forge/zeopp-lsmo)
Befor run this class please test "network" commond is works or not.
"""

import os
import subprocess

def ChanDim(structure, probe_radius = 0, high_accuracy = True, prefix="tmp_chan"):

    """Analysis dimension of channel.

    Args:
        structure (str): path to your CIF.
        probe_radius (float): probe of radiu.
        high_accuracy (bool): use high accuracy or not.
        prefix (str): temporary file.

    Returns:
        Dictionary:
            -   unit by ["unit"], always nan.
            -   dimention by ["Dimension"].
    """
    
    results_chan = {}
    results_chan["unit"]="nan"
    
    tmp_file = f"{prefix}.txt"
    
    if high_accuracy:
        cmd = f'network -ha -chan {probe_radius} {tmp_file} {structure}'
        # cmd = f'network -ha S50 -chan {probe_radius} {tmp_file} {structure}'
    else:
        cmd = f'network -chan {probe_radius} {tmp_file} {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )

    with open(tmp_file) as f:
        for i, row in enumerate(f):
            if i == 0:
                dim = int(row.split('dimensionality')[1].split()[0])

    results_chan["Dimension"] = dim

    os.remove(tmp_file)

    return results_chan

def FrameworkDim(structure, high_accuracy = True, prefix="tmp_strinfo"):

    """Analysis dimension of framework.

    Args:
        structure (str): path to your CIF.
        high_accuracy (bool): use high accuracy or not.
        prefix (str): temporary file.

    Returns:
        Dictionary:
            -   unit by ["unit"], always nan
            -   dimention by ["Dimension"]
            -   number of 2D framewor by ["N_1D"]
            -   number of 1D framework by ["N_2D"]
            -   number of 3D framewor by ["N_3D"]
    """

    results_strinfo = {}
    results_strinfo["unit"]="nan"

    tmp_file = f"{prefix}.txt"
    
    if high_accuracy:
        cmd = f'network -ha -strinfo {tmp_file} {structure}'
        # cmd = f'network -ha S50 -strinfo {tmp_file} {structure}'
    else:
        cmd = f'network -strinfo {tmp_file} {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )

    with open(tmp_file) as f:
        line = f.readline().split()
        try:
            dim = int(line[-1])
            one_dim = int(line[7])
            two_dim = int(line[8])
            three_dim = int(line[9])
        except:
            one_dim = 0
            two_dim = 0
            three_dim = 0
            dim = 0
    results_strinfo["Dimension"] = dim
    results_strinfo["N_1D"] = one_dim
    results_strinfo["N_2D"] = two_dim
    results_strinfo["N_3D"] = three_dim

    os.remove(tmp_file)

    return results_strinfo

def PoreDiameter(structure, high_accuracy = True, prefix="tmp_pd"):

    """Analysis pore diameter of structure.

    Args:
        structure (str): path to your CIF.
        high_accuracy (bool): use high accuracy or not.
        prefix (str): temporary file.

    Returns:
        Dictionary:
            -   unit by ["unit"], always angstrom, Å
            -   largest cavity diameter by ["LCD"]
            -   pore-limiting diameter by ["PLD"]
            -   largest free pore diameter by ["LFPD"]
    """

    results_pd = {}
    results_pd["unit"]="angstrom, Å"

    tmp_file = f"{prefix}.txt"
    
    if high_accuracy:
        cmd = f'network -ha -res {tmp_file} {structure}'
        # cmd = f'network -ha S50 -res {tmp_file} {structure}'
    else:
        cmd = f'network -res {tmp_file} {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )
    with open(tmp_file) as f:
        line = f.readline().split()
        results_pd["LCD"], results_pd["PLD"], results_pd["LFPD"] = map(float, line[1:4])
    os.remove(tmp_file)

    return results_pd

def SurfaceArea(structure, chan_radius = 1.655, probe_radius = 1.655, num_samples = 5000, high_accuracy = True, prefix="tmp_sa"):

    """Analysis surface area of structure.

    Args:
        structure (str): path to your CIF.
        chan_radius (float): probe of channel, it is advised to keep chan_radius=probe_radius.
        probe_radius (float): probe of radiu.
        num_samples (int): number of MC samples per atom.
        high_accuracy (bool): use high accuracy or not.
        prefix (str): temporary file.

    Returns:
        Dictionary:
            -   unit by ["unit"], always Å^2, m^2/cm^3, m^2/g
            -   accessible surface area by ["ASA"]
            -   non-accessible surface area by ["NASA"]
    """

    results_sa = {}
    results_sa["unit"]="Å^2, m^2/cm^3, m^2/g"

    tmp_file = f"{prefix}.txt"
    
    if high_accuracy:
        cmd = f'network -ha -sa {chan_radius} {probe_radius} {num_samples} {tmp_file} {structure}'
        # cmd = f'network -ha S50 -sa {chan_radius} {probe_radius} {num_samples} {tmp_file} {structure}'
    else:
        cmd = f'network -sa {chan_radius} {probe_radius} {num_samples} {tmp_file} {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )
    with open(tmp_file) as f:
        for i, row in enumerate(f):
            if i == 0:
                ASA = float(row.split('ASA_A^2:')[1].split()[0])
                VSA = float(row.split('ASA_m^2/cm^3:')[1].split()[0])
                GSA = float(row.split('ASA_m^2/g:')[1].split()[0])
                NASA = float(row.split('NASA_A^2:')[1].split()[0])
                NVSA = float(row.split('NASA_m^2/cm^3:')[1].split()[0])
                NGSA = float(row.split('NASA_m^2/g:')[1].split()[0])

    results_sa["ASA"] = [ASA, VSA, GSA]
    results_sa["NASA"] = [NASA, NVSA, NGSA]

    os.remove(tmp_file)

    return results_sa

def PoreVolume(structure, chan_radius = 0, probe_radius = 0, num_samples = 5000, high_accuracy = True, prefix="tmp_pv"):

    """Analysis pore volume of structure.

    Args:
        structure (str): path to your CIF.
        chan_radius (float): probe of channel, it is advised to keep chan_radius=probe_radius.
        probe_radius (float): probe of radiu.
        num_samples (int): number of MC samples per atom.
        high_accuracy (bool): use high accuracy or not.
        prefix (str): temporary file.

    Returns:
        Dictionary:
            -   unit by ["unit"], always PV: Å^3, cm^3/g; VF: nan
            -   accessible pore volume by ["PV"]
            -   non-accessible pore volume by ["NPV"]
            -   accessible void fraction by ["VF"]
            -   non-accessible void fraction by ["NVF"]
    """

    results_pv = {}
    results_pv["unit"]="PV: Å^3, cm^3/g; VF: nan"

    tmp_file = f"{prefix}.txt"
    
    if high_accuracy:
        cmd = f'network -ha -volpo {chan_radius} {probe_radius} {num_samples} {tmp_file} {structure}'
        # cmd = f'network -ha S50 -volpo {chan_radius} {probe_radius} {num_samples} {tmp_file} {structure}'
    else:
        cmd = f'network -volpo {chan_radius} {probe_radius} {num_samples} {tmp_file} {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )
    with open(tmp_file) as f:
        for i, row in enumerate(f):
            if i == 0:
                POAV = float(row.split('POAV_A^3:')[1].split()[0])
                PONAV = float(row.split('PONAV_A^3:')[1].split()[0])
                GPOAV = float(row.split('POAV_cm^3/g:')[1].split()[0])
                GPONAV = float(row.split('PONAV_cm^3/g:')[1].split()[0])
                POAV_volume_fraction = float(row.split('POAV_Volume_fraction:')[1].split()[0])
                PONAV_volume_fraction = float(row.split('PONAV_Volume_fraction:')[1].split()[0])
    results_pv["PV"] = [POAV, GPOAV]
    results_pv["NPV"] = [PONAV, GPONAV]
    results_pv["VF"] = POAV_volume_fraction
    results_pv["NVF"] = PONAV_volume_fraction

    os.remove(tmp_file)

    return results_pv



"""Get publication information from DOI.
"""

import requests, time


def get_publication_date(doi, max_retries=3, delay=5):
    """Get publicated date.

    Args:
        doi (str): DOI.
       
    Returns:
        str:
            -   time of DOI.
    """
        
    url = f"https://api.crossref.org/works/{doi}"
    for attempt in range(max_retries):
        try:
            response = requests.get(url, timeout=15)
            if response.status_code == 200:
                data = response.json().get("message", {})
                if 'published-online' in data:
                    date = data['published-online']['date-parts'][0]
                elif 'published-print' in data:
                    date = data['published-print']['date-parts'][0]
                elif 'license' in data:
                    date = data['license'][0]['start']['date-parts'][0]
                else:
                    return "unknown"
                if len(date) == 3:
                    return f"{date[0]}-{date[1]:02d}-{date[2]:02d}"
                elif len(date) == 2:
                    return f"{date[0]}-{date[1]:02d}"
                else:
                    return str(date[0])
            else:
                return "unknown"
        except requests.exceptions.ReadTimeout:
            print(f"[Timeout] DOI: {doi}, retrying ({attempt + 1}/{max_retries})...")
            time.sleep(delay)
        except Exception as e:
            print(f"[Error] DOI: {doi} failed due to {e}")
            return "unknown"
    return "unknown"


def extract_publication(doi):
    """Get publisher.

    Args:
        doi (str): DOI.
       
    Returns:
        str:
            -   Publisher
    """
    try:
        doi_part1 = doi.split("/")[0]
        part_1 = ["10.1021","10.1039","10.1002","10.1038","10.1126","10.1016","10.1007","10.3390"]
        part_name = ["ACS","RSC","WILEY","Nature","Science","SciDirect","Springer","MDPI"]
        try:
            index = part_1.index(doi_part1)
            return part_name[index]
        except:
            if doi_part1 == "unknown":
                return "unknown"
            else:
                return "other"
    except:
        return "unknown"